finnhub_exchanges_url: str = "https://raw.githubusercontent.com/amor71/LiuAlgoTrader/4520081c56aef7a23b37416ae6e95465b5e50f80/static/finnhub_exchanges.csv"
finnhub_indices_url: str = "https://raw.githubusercontent.com/amor71/LiuAlgoTrader/4520081c56aef7a23b37416ae6e95465b5e50f80/static/finnhub_indices.csv"
